//-----------------------------------------------------------------------------
// local includes
#include "Objects.h"
#include "Scene.h"

#include <math.h>


//-----------------------------------------------------------------------------
using namespace std;


//-----------------------------------------------------------------------------
void CObject::changeReflection( double delta )
{
	reflect += delta;
	if ( reflect < 0.0 ) reflect = 0.0;
	if ( reflect > 1.0 ) reflect = 1.0;
}

void CObject::changeRefraction( double delta )
{
	refract += delta;
	if ( refract < 0.0 ) refract = 0.0;
	if ( refract > 1.0 ) refract = 1.0;
}


//-----------------------------------------------------------------------------
const Point3d CSphere::intersection( const CRay& ray ) const
{
  // local coordinate
  const Point3d& v = p - ray.start();

  double t = v * ray.dir();
  const double v2 = v*v;
  const double d2 = v2 - t*t;
  
  // check whether sphere is behind us or we don't hit it
  if ( ( t < 0 ) || ( d2 > r2 ) )
    return Point3d( -numeric_limits<double>::max() );
     
  // check if we the ray starts inside the sphere
  if ( v2 >= r2 )
    t -= sqrt( r2 - d2 );
  else
    t += sqrt( r2 - d2 );
 
  // point of intersection
  return ray.start() + t * ray.dir();
}


//-----------------------------------------------------------------------------
const Point3d CSphere::normalAt( const Point3d& iP ) const
{
  return ( iP - pos() ).normalized();
}


//-----------------------------------------------------------------------------
const CColor CSphere::colorAt( const Point3d&, const Point3d& n, 
                               const Point3d& l, const CLight& light ) const
{
  // light coefficient
  const double cos_phi = l * n;

  // backside
  if ( cos_phi < 0 ) 
    return CColor();
    
  // return the color
  return cos_phi * ( light.color() * c );
}


//-----------------------------------------------------------------------------
const Point3d CPlane::intersection( const CRay& ray ) const
{
  // distance to the plane
  const double dist = ray.start() * normal - d;

  // parallel?
  double p1n = ray.dir() * normal;
  if ( fabs( p1n ) < numeric_limits<double>::epsilon() )
    return Point3d( -numeric_limits<double>::max() );

  // check if the ray intersects the plane
  if ( ( dist > 0 && p1n > 0 ) || ( dist < 0 && p1n < 0 ) )
    return Point3d( -numeric_limits<double>::max() );

  // return the intersection point
  return ray.start() + ( ray.dir() * ( -dist / p1n ) );
}


//-----------------------------------------------------------------------------
const Point3d CPlane::normalAt( const Point3d& ) const
{
  return normal;
}


//-----------------------------------------------------------------------------
const CColor CPlane::colorAt( const Point3d& iP, const Point3d& n,
                              const Point3d& l, const CLight& light ) const
{
  // diffuse coefficient
  const double cos_phi = l * n;

  // backside
  if ( cos_phi < 0 )
    return CColor();
    
  // return the color (chess-board colored)
  const int x = iP.x() >= 0? (int)iP.x() : (int)iP.x() - 1;
  const int z = iP.z() >= 0? (int)iP.z() : (int)iP.z() - 1;
  if ( abs( x % 2 ) == abs( z % 2 ) )
    return cos_phi * ( light.color() * c );
  
  return cos_phi * ( light.color() * CColor( 255, 255, 255 ) );
}
