//-----------------------------------------------------------------------------
// local includes
#include "Scene.h"
#include "Objects.h"


//-----------------------------------------------------------------------------
using namespace std;


//-----------------------------------------------------------------------------
CScene::~CScene()
{
  // delete the objects
  for ( unsigned i = 0; i < objects.size(); i++ )
    delete objects[ i ];

  // delete the lights
  for ( unsigned i = 0; i < lights.size(); i++ )
    delete lights[ i ];
}


//-----------------------------------------------------------------------------
const bool CScene::intersectionTest( const CRay& ray ) const
{
  // test if the ray intersects an object
  for ( unsigned int i = 0; i < numObjects(); i++ )
  {
    // self check
    if ( objects[ i ] == ray.source() )
      continue;

    // calculate intersection point
    const Point3d& actIP = objects[ i ]->intersection( ray );
    
    // check if an intersection point exists
    if ( actIP.z() > nearC && actIP.z() < farC )
      return true;
  }

  return false;
}


//-----------------------------------------------------------------------------
const CColor CScene::intersection( const CRay& ray, const int iter ) const
{
  // search the closest object
  CObject* obj = NULL;
  double dist = numeric_limits<double>::max();
  Point3d iP;
  for ( unsigned int i = 0; i < numObjects(); i++ )
  {
    // calculate intersection point
    const Point3d& actIP = objects[ i ]->intersection( ray );
    
    // check if an intersection point exists
    if ( actIP.z() <= nearC || actIP.z() >= farC )
      continue;

    // get the distance to the intersection point
    const double actDist = ( ray.start() - actIP ).squaredNorm();
    if ( actDist >= dist )
      continue;

    // set the intersection point and the object
    iP = actIP;
    dist = actDist;
    obj = objects[ i ];
  }

  // check for intersection
  if ( obj == NULL )
  {
    if ( iter == 0 || ( ray.source() && ray.source()->refraction() > 0 ) )
      return backgroundColor();
    return CColor();
  }

  // return the color
  return shading( obj, ray, iP, iter );
}


//-----------------------------------------------------------------------------
const CColor CScene::shading( const CObject* obj, const CRay& ray, 
                              const Point3d& iP, const int iter ) const
{
  // object color at intersection point
  CColor color;

  // get the normal of the object at the intersection point
  const Point3d& normal = obj->normalAt( iP );

  // sum up diffuse color for each light
  for ( int i = 0; i < numLights(); i++ )
  {
    // get the light
    const CLight& actLight = light( i );

    // light direction vector
    const Point3d& l = ( actLight.pos() - iP ).normalized();

    // ray from intersection point to light
    CRay lightVec( iP, l, obj );

    // check if an object lies between the light an this object
    if ( intersectionTest( lightVec ) )
      continue;

    // add the color
    color += obj->colorAt( iP, normal, l, actLight );
  }

  // add the reflected color
  if ( iter < maxDepth && obj->reflection() > 0 )
  {
    // secundary ray
    const Point3d& sec = ray.dir() - 2 * ( ray.dir() * normal ) * normal;

    // add the color
    color += obj->reflection() 
      * intersection( CRay( iP + sec*0.0001, sec, obj ), iter+1 );
  }

  // add the refracted color
  if ( iter < maxDepth && obj->refraction() > 0 )
  {
    // refraction index (air->glass)
    double refract = 1 / 1.7;
    // refraction index (glass->air)
    if ( obj == ray.source() )
      refract = 1.7;

    // invert the normal if we are inside an object
    Point3d n = normal;
    if ( obj == ray.source() )
      n *= -1;

    // cos of incoming angle
    const double cosTheta1 = n * -ray.dir();
    // sin of incoming angle
    const double sinTheta1 = sin( acos( cosTheta1 ) );
    // cos of outgoing angle
    const double h0 = refract * sinTheta1;
    const double cosTheta2 = sqrt( 1 - h0*h0 );

    // refracted ray
    const Point3d& t = -n*cosTheta2 + refract*( ray.dir() + n*cosTheta1 );

    // trace the refracted ray
    color += obj->refraction() 
      * intersection( CRay( iP + t*0.0001, t, obj ), iter+1 ) * obj->color();
  }

  // return the color
  return color.normalize();
}
